/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  field.cpp
///  \brief  Implementation of finite fields.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   06.10.2010
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "field.hpp"
#include "helpfunction.hpp"

unsigned int Field::m_p;
unsigned int Field::m_r;
unsigned int Field::m_q;

INDEX Field::m_alpha;

std::vector<INDEX> Field::m_primeFieldElements;
std::vector<INDEX> Field::m_addition;
std::vector<INDEX> Field::m_subtraction;
std::vector<INDEX> Field::m_multiplication;
std::vector<INDEX> Field::m_division;
std::vector<INDEX> Field::m_additive_inv;
std::vector<INDEX> Field::m_multiplicative_inv;
std::vector<std::vector<INDEX> > Field::m_frobenius;
std::vector<std::vector<INDEX> > Field::m_basis_representation;
unsigned int Field::m_spacing;
unsigned int Field::m_bitsPerEntry;

void Field::setSpacing_bitsPerEntry(void)
{
    m_spacing = 1;
    unsigned int y = 10;
    while(y<m_q)
    {
        ++m_spacing;
        y *= 10;
    }
    ++m_spacing;

    y = 2;
    m_bitsPerEntry = 1;
    while(y<m_q)
    {
        m_bitsPerEntry++;
        y<<=1;
    }
}

void Field::init(unsigned int p)
{
	m_p = p;
	m_r = 1;
	m_q = p;
	setSpacing_bitsPerEntry();


	for(m_alpha=2; m_alpha<p; ++m_alpha)
	{
		bool isGen=true;
		INDEX pow = (m_alpha*m_alpha)%p;
		for(unsigned int exp=2; exp<p-1; ++exp, pow = (pow * m_alpha)%p)
		{
			if(pow == 1)
			{
				isGen =false;
				exp =p;
			}
		}
		if(isGen)
		{
			break;
		}
	}

	unsigned int squared_p = p*p;
	m_primeFieldElements.resize(p,0);
	m_addition.resize(squared_p);
	m_subtraction.resize(squared_p);
	m_multiplication.resize(squared_p);
	m_division.resize(squared_p);
	m_additive_inv.resize(p, 0);
	m_multiplicative_inv.resize(p, 0);
	m_frobenius.clear();
	m_basis_representation = std::vector<std::vector<INDEX> >(m_p, std::vector<INDEX>(1,0));

	std::vector<INDEX>::iterator addIter(m_addition.begin());
	std::vector<INDEX>::iterator subtractIter(m_subtraction.begin());
	std::vector<INDEX>::iterator multIter(m_multiplication.begin());
	std::vector<INDEX>::iterator divIter(m_division.begin());
	std::vector<INDEX>::iterator addInvIter(m_additive_inv.begin() + 1);
	std::vector<INDEX>::iterator multInvIter(m_multiplicative_inv.begin() +1);
//	std::vector<INDEX>::iterator frobeniusIter(m_frobenius[0].begin() +1);

	for(INDEX a=1; a<p; ++a)
	{
	    m_primeFieldElements[a] = a;
		*addInvIter = p-a;
		++addInvIter;

		// the multiplicative inverse:
		INDEX pow(a);
		while( (pow*a)%p  != 1)
		{
			pow *= a;
			pow %=p;
		}
		*multInvIter =pow;
		++multInvIter;
//
//		*frobeniusIter = a;
//		++frobeniusIter;
		m_basis_representation[a][0] = a;
	}
	for(INDEX a=0; a<p; ++a)
	{
		for(INDEX b=0; b<p; ++b)
		{
			*addIter = (a+b)%p;
			++addIter;
			*subtractIter = (a - b + p)%p;
			++subtractIter;
			*multIter = (a*b)%p;
			++multIter;
			*divIter = (a * m_multiplicative_inv[b])%p;
			++divIter;
		}
	}
}

std::ostream& Field::print(std::ostream& os)
{
	std::vector<INDEX>::iterator addInvIter(m_additive_inv.begin());
	std::vector<INDEX>::iterator addInvIterEnd(m_additive_inv.end());
	os<<"additive Inverse"<<std::endl;
	for( ; addInvIter<addInvIterEnd; ++addInvIter)
	{
		os<<*addInvIter<<" ";
	}
	os<<std::endl<<std::endl;

	std::vector<INDEX>::iterator multInvIter(m_multiplicative_inv.begin());
	std::vector<INDEX>::iterator multInvIterEnd(m_multiplicative_inv.end());
	os<<"multiplicative Inverse"<<std::endl;
	for( ; multInvIter<multInvIterEnd; ++multInvIter)
	{
		os<<*multInvIter<<" ";
	}
	os<<std::endl<<std::endl;

	if(m_r>1)
	{
	    os<<"Frobenius images"<<std::endl;
	    for(unsigned int exp=1; exp<m_r; ++exp)
	    {
            std::vector<INDEX>::iterator frobIter(m_frobenius[exp].begin() );
            std::vector<INDEX>::iterator frobIterEnd(m_frobenius[exp].end() );

            for( ; frobIter<frobIterEnd; ++frobIter)
            {
                os<<*frobIter<<" ";
            }
            os<<std::endl;
	    }
	}
	os<<std::endl;

	std::vector<INDEX>::iterator addIter(m_addition.begin());
	os<<"Addition table"<<std::endl;
	for(INDEX a=0; a<m_q; ++a)
	{
		for(INDEX b=0; b<m_q; ++b)
		{
			os<<*addIter<<" ";
			++addIter;
		}
		os<<std::endl;;
	}

	std::vector<INDEX>::iterator subtractIter(m_subtraction.begin());
	os<<std::endl<<std::endl;
	os<<"Subtraction table"<<std::endl;
	for(INDEX a=0; a<m_q; ++a)
	{
		for(INDEX b=0; b<m_q; ++b)
		{
			os<<*subtractIter<<" ";
			++subtractIter;
		}
		os<<std::endl;;
	}

	std::vector<INDEX>::iterator multIter(m_multiplication.begin());
	os<<std::endl<<std::endl;
	os<<"Multiplication table"<<std::endl;
	for(INDEX a=0; a<m_q; ++a)
	{
		for(INDEX b=0; b<m_q; ++b)
		{
			os<<*multIter<<" ";
			++multIter;
		}
		os<<std::endl;;
	}

	std::vector<INDEX>::iterator divIter(m_division.begin());
	os<<std::endl<<std::endl;
	os<<"Division table"<<std::endl;
	for(INDEX a=0; a<m_q; ++a)
	{
		for(INDEX b=0; b<m_q; ++b)
		{
			os<<*divIter<<" ";
			++divIter;
		}
		os<<std::endl;;
	}
	return os;
}

void Field::init(const char* const filename)
{
	std::ifstream in;
	in.open(filename);

	if(!in) {
		std::cerr<<"Unable to open "<<filename<<std::endl;
		exit(-1);
	}

	in>>m_p;
	in>>m_r;
	in>>m_q;
	setSpacing_bitsPerEntry();

	unsigned int squared_q = m_q * m_q;
	m_primeFieldElements.resize(m_p);
	m_addition.resize(squared_q);
	m_subtraction.resize(squared_q);
	m_multiplication.resize(squared_q);
	m_division.resize(squared_q);
	m_additive_inv.resize(m_q);
	m_multiplicative_inv.resize(m_q);
    m_basis_representation = std::vector<std::vector<INDEX> >(m_q, std::vector<INDEX>(m_r,0));

	std::vector<INDEX>::iterator addIter(m_addition.begin());

	for(unsigned int a=0; a<m_q; ++a)
	{
		for(unsigned int b=0; b<m_q; ++b)
		{
			in >> *addIter;
			++addIter;
		}
	}

	std::vector<INDEX>::iterator multIter(m_multiplication.begin());
	for(unsigned int a=0; a<m_q; ++a)
	{
		for(unsigned int b=0; b<m_q; ++b)
		{
			in >> *multIter;
			++multIter;
		}
	}

	std::vector<INDEX>::iterator addInvIterBeg(m_additive_inv.begin());
	std::vector<INDEX>::iterator addInvIterEnd(m_additive_inv.end());
	std::vector<INDEX>::iterator addInvIter(addInvIterBeg);
	for( ;addInvIter<addInvIterEnd; ++addInvIter )
	{
		in >> *addInvIter;
	}

	std::vector<INDEX>::iterator multInvIterBeg(m_multiplicative_inv.begin());
	std::vector<INDEX>::iterator multInvIterEnd(m_multiplicative_inv.end());
	std::vector<INDEX>::iterator multInvIter(multInvIterBeg);
	for( ; multInvIter<multInvIterEnd ; ++multInvIter )
	{
		in >> *multInvIter;
	}

	in >> m_alpha;

	if(m_r>1)
	{ // not needed for prime fields
		m_frobenius.resize(m_q);
		m_frobenius[1].resize(m_q);
		std::vector<unsigned int>& frobImg = m_frobenius[1];
		std::vector<INDEX>::iterator frobIter(frobImg.begin() );
		std::vector<INDEX>::iterator frobIterEnd(frobImg.end() );
		for( ; frobIter<frobIterEnd; ++frobIter)
		{
			in >> *frobIter;
		}

		for(unsigned int exp= 2; exp<m_r; ++exp)
		{
			m_frobenius[exp].resize(m_q);
			std::vector<INDEX>::iterator frobPredIter(m_frobenius[exp-1].begin() );
			std::vector<INDEX>::iterator frobPowIter(m_frobenius[exp].begin() );
			std::vector<INDEX>::iterator frobPowIterEnd(m_frobenius[exp].end() );
			for( ; frobPowIter<frobPowIterEnd; ++frobPowIter, ++frobPredIter)
			{
				*frobPowIter =  frobImg[ *frobPredIter ];
			}
		}

		std::vector<std::vector<INDEX> >::iterator valIter(m_basis_representation.begin());
		for(unsigned int a=0; a<m_q; ++a, ++valIter)
		{
		    std::vector<INDEX>::iterator basisIter(valIter->begin());
		    for(unsigned int exp=0; exp<m_r; ++exp, ++basisIter)
		    {
		        in>>*basisIter;
		    }
		}
	}
	else
	{
	    std::vector<std::vector<INDEX> >::iterator valIter(m_basis_representation.begin());
        for(unsigned int a=0; a<m_q; ++a, ++valIter)
        {
            valIter->front() = a;
        }
	}

	// fill subtraction
	std::vector<INDEX>::iterator subtractIter(m_subtraction.begin());
	for(unsigned int a=0; a<m_q; ++a)
	{
		for(addInvIter = addInvIterBeg; addInvIter<addInvIterEnd ; ++addInvIter, ++subtractIter)
		{
			*subtractIter = m_addition[a*m_q + (*addInvIter)];
		}
	}

	// fill division
	std::vector<INDEX>::iterator divIter(m_division.begin());
	for(unsigned int a=0; a<m_q; ++a)
	{
		for(multInvIter = multInvIterBeg; multInvIter<multInvIterEnd ; ++multInvIter, ++divIter)
		{
			*divIter = m_multiplication[a*m_q + (*multInvIter)];
		}
	}
	Element<Field> res(0);
	for(unsigned int i=1; i<m_p; i++)
	{
	    res+=1;
	    m_primeFieldElements[i] = res.getIndex();
	}

	in.close();
}

std::vector<Element<Field> > Field::nontrivalUnitsPlusMinus(void)
{
    std::vector<Element<Field> > nontrivialUnits(Field::nontrivalUnits());
    std::vector<Element<Field> >::const_iterator nontrivialUnitsIterEnd(nontrivialUnits.end());
    std::vector<Element<Field> > nontrivialUnitsPlusMinus( (nontrivialUnits.size()-1) >>1 ); // will not return {+-1}
    std::vector<Element<Field> >::iterator nontrivialUnitsPlusMinusIter(nontrivialUnitsPlusMinus.begin());

    for( std::vector<Element<Field> >::const_iterator nontrivialUnitsIter(nontrivialUnits.begin()); nontrivialUnitsIter < nontrivialUnitsIterEnd; ++nontrivialUnitsIter)
    {
        if(*nontrivialUnitsIter< -(*nontrivialUnitsIter))
        {
            *nontrivialUnitsPlusMinusIter = *nontrivialUnitsIter;
            ++nontrivialUnitsPlusMinusIter;
        }
    }

    return nontrivialUnitsPlusMinus;
}
