/* This is  -*- C++ -*-  */
// *********************************************************************
// **  groups.cpp
// *********************************************************************
// **
// **
// **
// **  Copyright (C) 2010 Ralf Gugisch
// ** 	Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
// **  http://www.gnu.org/licenses/
// *********************************************************************

#include <assert.h>
#include <stdarg.h>
#include <queue>
#include <stack>
#include <set>
#include<sstream>

#include "groups.h"

#include <iomanip> // for operator<<


// *****************************************************************************
// ** class Permutation
// *****************************************************************************

Permutation::PermNotation Permutation::print_style= Permutation::cycle_notation;
//Permutation::PermNotation Permutation::print_style= Permutation::list_notation;

Permutation cyclePermutation(int len, int a0,int a1,...) {
	Permutation ret;
	assert(len>=2);
	std::vector<int> P(len,-1);
	std::vector<int>::iterator it=P.begin();
	*it++=a0; *it++=a1;
	va_list a; va_start(a,a1);
	for(; it!=P.end(); ++it) *it=va_arg(a,int);
	va_end(a);
	return ret.cycle(P.begin(),P.end());
}
Permutation listPermutation(int len, int a0,int a1,...) {
	std::vector<int> P(len,-1);
	assert(len>=2);
	std::vector<int>::iterator it=P.begin();
	*it++=a0; *it++=a1;
	va_list a; va_start(a,a1);
	for(; it!=P.end(); ++it) *it=va_arg(a,int);
	va_end(a);
	asPermutation(P).trim();
	return asPermutation(P);
}
Permutation fillupPermutation(int len, int a0,int a1,...) {
	Permutation ret;
	assert(len>=2);
	std::vector<int> P(len,-1);
	std::vector<int>::iterator it=P.begin();
	*it++=a0; *it++=a1;
	va_list a; va_start(a,a1);
	for(; it!=P.end(); ++it) *it=va_arg(a,int);
	va_end(a);
	return ret.fillup(P.begin(),P.end());
}

//int Permutation::compare(Permutation const& p) const {
//	// TF: introduced maximum instead of calling degree in each step.
//	int maximum = max(degree(),p.degree());
//	for(int i=0; i<maximum; ++i)
//    if(apply(i)!=p[i]) return apply(i)-p[i];
//  return 0;
//}

int Permutation::compare(Permutation const& p) const {
	// TF: introduced maximum instead of calling degree in each step.
	int this_deg = degree(), p_deg = p.degree();
	for(int i=0; i<std::max(this_deg, p_deg); ++i)
		if(apply(i, this_deg)!=p.apply(i, p_deg)) return apply(i, this_deg)-p.apply(i, p_deg);
	return 0;
}

Permutation&  Permutation::random(int n) {
	int i;
	std::set<int> numbers;
	if(n) v.resize(n);
	for(i=degree()-1;i>=0;--i)  {
		int j=int(double((rand())/(RAND_MAX+1.0)*(i+1)));  // j \in [0;i]
		int n=0; while(numbers.find(n)!=numbers.end()) ++n; // n=first unused number
		while(j--) // n=next unused number
		{ ++n; while(numbers.find(n)!=numbers.end()) ++n; }
		v[i] = n;
		numbers.insert(n);
	}
	trim();

	return *this;
}

bool Permutation::valid() const {
	std::vector<int> used(v);
	std::sort(used.begin(),used.end());
	// Test, if given vector contains negative elements
	if(used[0]<0) return false;
	// Test, if given vector contains no duplicates
	if(std::unique(used.begin(),used.end())!=used.end()) return false;
	// Test, if given vector has gaps
	if(used.back()-used.size()+1!=0) return false;
	return true;
}

// ******************************************

bool Permutation::sign() const {
	int i,j,x=0,n=degree();
	for(i=0;i<n;i++) for(j=i+1;j<n;j++) if(apply(i, n)>apply(j,n)) ++x;
	return (x%2);
}

// ******************************************

std::string Permutation::LIST_SEP = " ";

void Permutation::print_l(std::ostream& out, unsigned int perm_offset) const {
	std::ostringstream x;
	int i, deg = degree();
	bool first=true;
	x << "[";
	for(i=0; i<deg; ++i) {
		if(first) first=false; else x<<LIST_SEP; // ' ' is easier for scanning
		x << std::setw(2) << apply(i, deg) + perm_offset;
	}
	x <<"]";
	out << x.str();
}

void Permutation::magma_print(std::ostream& out, int n) const {
	std::ostringstream x;
	int i, deg = degree();
	bool first=true;
	x << "[";
	// degree() <= n is necessary!
	if(deg>n)
	{
		std::cerr<<"The argument n in Permutation::magma_print is wrong"<<std::endl;
		n = deg;
	}

	for(i=0; i<n; ++i) {
		if(first) first=false; else x<<", "; // ', ' is easier for magma
		x << std::setw(2) << apply(i, deg) + 1;
	}
	x <<"]";
	out << x.str();
}


void Permutation::print_c(std::ostream& out,  unsigned int perm_offset) const {
	std::ostringstream x;
	int i,j, deg = degree();
	std::vector<bool> finished(degree(),false);
	bool id=true;
	for(i=0; i<deg; i++) {
		if(finished[i]) continue;
		finished[i]=true;
		// new cycle
		j=apply(i, deg);
		if(j==i) continue; // fix point
		id=false; // this is _not_ the identity
		x<< "(" << /*std::setw(2) <<*/ i +perm_offset;
		for(;j!=i; j=apply(j, deg)) {
			finished[j]=true;
			x<< "," << /*std::setw(2) <<*/ j +perm_offset;
		}
		x <<")";
	}
	if (id) x << "(1)";
	out << x.str();
}

// ******************************************

bool Permutation::scan_l(std::istream& in, unsigned int perm_offset) {
	// scan permutation as second-row list [...]
	int x;
	int dim=0;
	std::vector<int> v;
	in>>std::ws;
	if(in.peek()!='[') return false;
	in.ignore(); in>>std::ws;
	while(in && in.peek()!=']') {
		in >> x >> std::ws;
		x=x-perm_offset;
		if(in.peek()==',') { in.ignore(); in>>std::ws; }
		v.push_back(x); if(x+1>dim) dim=x+1;
	}
	in.ignore(); // remove the ']'
	if((unsigned int)dim!=v.size()) return false;
	list(v);
	return true; // not implemented
}

bool Permutation::scan_c(std::istream& in, unsigned int perm_offset) {
	// scan permutation as cycle-list (...)...(...)
	
	int dim=0, x,x0;
	std::vector<int> v;
	in>>std::ws;
	if(in.peek()!='(') return false;

	while(in && in.peek()=='(') {
		in.ignore(); in>>std::ws;
		while(in && in.peek()!=')') {
			in >> x >> std::ws;
			x=x-perm_offset;
			if(in.peek()==',') { in.ignore(); in>>std::ws; }
			v.push_back(x); if(x+1>dim) dim=x+1;
		}
		if(!in) return false;
		in.ignore();
		v.push_back(-1);
	}
	std::vector<int> p0(dim);
	for(int i=0; i<dim; ++i) 
		p0[i]=i;
	for(unsigned int i=0; i<v.size();++i) {
		x0=x=v[i];
		for(++i; v[i]!=-1;++i) {
			if(p0[x]!=x) return false;
			p0[x]=v[i]; x=v[i];
		}
		if(p0[x]!=x) return false;
		p0[x]=x0;
	}
	this->v.swap(p0);
	return true;
}

// ******************************************

/*/// \relates Permutation
/// \a ret will be set on \f$ A \circ B \f$.
///  It is explicitely allowed to overwrite arguments.
///  (In case of overwriting \a A, a local copy is created.)
/// von RALF
Permutation& mul(Permutation const& A, Permutation const& B, Permutation &ret){
  if(A.isId()) { ret = B;  return ret;}
  if(B.isId()) { ret = A;  return ret;}

  if(&A==&ret && &B==&ret) { Permutation tmp; tmp.swap(ret); return mul(tmp,tmp,ret); }
  if(&A==&ret)             { Permutation tmp; tmp.swap(ret); return mul(tmp,  B,ret); }
  // &B==&ret is allowed
  int i=max(A.degree(),B.degree());
  while(--i>=0) if(i!=A[B[i]]) break; // skip trivial entries
//   ret.v.resize(++i,-1);
//   while(--i>=0) ret.v[i]=A[B[i]];
  // The following version avoids unnecessary initialization of vector:
  int j,s=ret.v.size();
  if(++i<=s) {
    ret.v.resize(i);
    for(j=0; j<i; ++j) ret.v[j]=A[B[j]];
  } else {
    ret.v.reserve(i);
    for(j=0; j<s; ++j) ret.v[j]=A[B[j]];
    for(;    j<i; ++j) ret.v.push_back(A[B[j]]);
  }
  return ret;
}*/


/// \relates Permutation
/// \a ret will be set on \f$ A \circ B \f$.
///  It is explicitely allowed to overwrite arguments.
///  (In case of overwriting \a A, a local copy is created.)
/// von THOMAS
Permutation& Permutation::mul(Permutation const& A, Permutation const& B, Permutation &ret){
	if(A.isId()) { ret = B;  return ret;}
	if(B.isId()) { ret = A;  return ret;}

	if(&A==&ret && &B==&ret) { Permutation tmp; tmp.swap(ret); return mul(tmp,tmp,ret); }
	if(&A==&ret)             { Permutation tmp; tmp.swap(ret); return mul(tmp,  B,ret); }
	// &B==&ret is allowed
	int B_deg = B.degree(), A_deg = A.degree(); 
	int i=std::max(A.degree(),B_deg);
	i--;
	while(i>=B_deg) {if(i!=A.apply(i, A_deg)) break; i--; } // skip trivial entries
	//   ret.v.resize(++i,-1);
	//   while(--i>=0) ret.v[i]=A[B[i]];
	// The following version avoids unnecessary initialization of vector:
	if(i==B_deg-1)
	{
		std::vector<int>::const_reverse_iterator B_rit = B.v.rbegin();
		while(i>=0) 
		{
			if(*B_rit==-1)
			{
				if(i!=A.apply(i, A_deg))
				{
					break;
				}
			}
			else
			{
				if(i!= A.apply(*B_rit, A_deg)) 
					break;
			}
			i--; 
			B_rit++; 
		} // skip trivial entries
	}

	int j;
	i++;
	ret.v.resize(i);
	std::vector<int>::iterator ret_it = ret.v.begin();
	std::vector<int>::const_iterator B_it = B.v.begin(), A_it;
	if(B_deg < i)
	{
		for(j=0; j<B_deg; ++j)
		{
			if(*B_it==-1)
			{
				*ret_it =  A.apply(j, A_deg);
			}
			else
			{
				*ret_it = A.apply(*B_it, A_deg);
			}
			ret_it++;
			B_it++;
		}
		A_it = A.v.begin()+B_deg;
		for(j=B_deg; j<i; ++j)
		{
			*ret_it = *A_it;
			A_it++;
			ret_it++;
		}
	}
	else
	{
		for(j=0; j<i; ++j)
		{
			if(*B_it==-1)
			{
				*ret_it = A.apply(j, A_deg);
			}
			else
			{
				*ret_it = A.apply(*B_it, A_deg);
			}
			ret_it++;
			B_it++;
		}
	}
	return ret;
}

Permutation& Permutation::mulTransposition(int t_min, int t_max, Permutation & A)
{
	if(t_min != t_max)
	{		
		int pre_min, pre_max, A_deg = A.degree();
		pre_min = A.invapply(t_min, A_deg);
		pre_max = A.invapply(t_max, A_deg);

		if(A_deg < t_max+1)
		{
			A.v.resize(t_max+1, -1);
		}

		A.v[pre_min] = t_max;
		A.v[pre_max] = t_min;

		if(pre_min == t_max && t_max == A_deg-1)
		{
			A.v.pop_back(); //The last point is t_max and mapped to itself
			A.trim();
		}
	}

	return A;
}	

Permutation& Permutation::mulTransposition(Permutation const& A, int t_min, int t_max, Permutation& ret)
{
	int max_ind = A.degree();
	if(A.isId()) { ret.transposition(t_min, t_max); return ret; }
	if(max_ind > t_max)
	{
		if(&ret != &A)
		{
			ret = A;
		}
		int dummy = A.v[t_min];
		ret.v[t_min] = A.v[t_max];
		ret.v[t_max] = dummy;
		if(max_ind == t_max+1 && dummy == t_max)
		{
			ret.trim();
		}
	}
	else
	{
		ret.v.resize(t_max+1);
		if(&ret != &A)
		{
			for(int i=0; i<max_ind; i++)
			{
				ret.v[i] = A.v[i];
			}
		}
		for(int i=max_ind; i<t_max; i++)
		{
			ret.v[i] = i;
		}
		ret.v[t_max] = ret.v[t_min];
		ret.v[t_min] = t_max;
	}
	return ret;
}

/// \relates Permutation
/// \a ret will be set on \f$ A \circ B^{-1} \f$.
///  It is explicitely allowed to overwrite arguments. (Local copies are created.)
Permutation& Permutation::div(Permutation const& A, Permutation const& B, Permutation &ret){
	if(&A==&ret && &B==&ret) { Permutation tmp; tmp.swap(ret); return div(tmp,tmp,ret); }
	if(&A==&ret)             { Permutation tmp; tmp.swap(ret); return div(tmp,  B,ret); }
	if(&B==&ret)             { Permutation tmp; tmp.swap(ret); return div(  A,tmp,ret); }
	int i=std::max(A.degree(),B.degree());
	ret.v.resize(i,-1);
	while(--i>=0) ret.v[B[i]]=A[i];
	ret.trim();
	return ret;
}

/// \relates Permutation
/// \a ret will be set on \f$ B \circ A \circ B^-1\f$.
///  It is explicitely allowed to overwrite arguments.
///  (In case of overwriting \a A, a local copy is created.)
Permutation& Permutation::pow(Permutation const& A, Permutation const& B, Permutation &ret){
	if(&A==&ret && &B==&ret) { Permutation tmp; tmp.swap(ret); return pow(tmp,tmp,ret); }
	if(&A==&ret)             { Permutation tmp; tmp.swap(ret); return pow(tmp,  B,ret); }
	if(&B==&ret)             { Permutation tmp; tmp.swap(ret); return pow(  A,tmp,ret); }
	int i=std::max(A.degree(),B.degree());
	ret.v.resize(i,-1);
	while(--i>=0) ret.v[B[i]]=B[A[i]];
	ret.trim();
	return ret;
}

/// \relates Permutation
/// \a ret will be set on \f$ A^{-1} \f$.
///  It is explicitely allowed to overwrite the argument. (A local copy is created.)
Permutation& Permutation::inv(Permutation const& A, Permutation &ret) {
	if(&A==&ret) return inv(Permutation(A),ret);
	int end=A.degree(), i=0;
	std::vector<int>::const_iterator A_v_it = A.getBegin();
	ret.v.resize(end);
	while(i<end)
	{
		if(*A_v_it != -1)
		{
			ret.v[*A_v_it]=i;
		}
		else
			ret.v[i] = i;
		i++;
		A_v_it++;
	}
	return ret;
}

Permutation& Permutation::cycle(unsigned int beg, unsigned int end)
{
	init(end+1);
	for(unsigned int i=beg; i<end;i++)
	{
		v[i] = i+1;
	}
	v[end] = beg;
	return *this;
}
