/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  innerStab.hpp
///  \brief Implements the stabilizer of the nodes of our backtrack tree.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#ifndef INNER_STAB_HPP_
#define INNER_STAB_HPP_

#include <iostream>
#include "partition.h"
#include "vector.hpp"
#include "field.hpp"
#include "defs.h"

template<typename RING, AlgTypeE ALGTYPE>
class ActingElement;

template<typename RING>
class Matrix;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class InnerStab
/// \brief Implements the inner stabilizer of a node, i.e. a subgroup of \f$ GL_k(R) \times {R^*}^n\f$
///
/// As long as the projection of the corresponding node generates a free module,
/// we use a representation of the stabilizer with the help of \a m_s0, \a m_s1 and \a row_partition.
/// If the projection is not a free module, we represent the generators by binary vectors of some length \a x.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
template<typename RING, AlgTypeE ALGTYPE>
struct InnerStab
{
};

template<typename RING>
struct InnerStab<RING, PERMUTATIONAL>
{
//    unsigned int* m_type;                                        ///< Tells you the type of the projection.
    unsigned int m_rank;                                        ///< The sum of the array m_type.
    std::vector<unsigned int> m_fixedCols;           ///< Tells the order in which the columns were fixed.
    std::vector<bool> m_isFixedCol;                        ///< Index i is true, if this column is fixed.

    InnerStab(void) : m_rank(0) {};
    /// Set the inner stabilizer to the group \f$ GL_k(R) \times {R^*}^n\f$.
    InnerStab(unsigned int k, unsigned int n) : m_rank(0),m_isFixedCol(n, false){}

    bool needUnit(unsigned int dummyVar) const { return false; }

    /// Minimize the last fixed column.
    void minimize(unsigned int k0, bool isInSpan, Matrix<RING>& matrix, ActingElement<RING, PERMUTATIONAL>* const transporter);

    /// Minimize the last fixed column in the case that the projection is a free module and that the next column is independent.
//    void minimizeIndependentK0(Matrix<RING>& matrix, ActingElement<RING, PERMUTATIONAL>* const transporter);

    /// There is no nontrivial inner stabilizer in this case. This method is needed as dummy call.
    My_Long_Int getAutom(const Matrix<RING>& matrix, std::vector<ActingElement<RING, PERMUTATIONAL> >& gens) const { return 1; }

    void fillActingElement(unsigned int k0, const Matrix<RING>& root,  ActingElement<RING, PERMUTATIONAL>& g,
            const std::vector<unsigned int>& to_fixSeq);

    friend std::ostream& operator<<(std::ostream& os, const InnerStab s)
    {
        os<<s.m_rank<<std::endl;
        return os;
    }
};

template<>
struct InnerStab<Field, LINEAR>
{
    unsigned int m_rank;								///< The sum of the array m_type.
    UnordPartition m_row_partition;				///< Sets of rows, which have to be mulitplied by the same unit.
    std::vector<unsigned int> m_fixedCols;			///< Tells the order in which the columns were fixed.
    std::vector<bool> m_isFixedCol;						///< Index i is true, if this column is fixed.

    InnerStab(void) : m_rank(0) {};
    /// Set the inner stabilizer to the group \f$ GL_k(R) \times {R^*}^n\f$.
    InnerStab(unsigned int k, unsigned int n) : m_rank(0), m_row_partition(k), m_isFixedCol(n, false){}

    bool needUnit(unsigned int dummyVar) const { return false; }

    /// Minimize the last fixed column.
    void minimize(unsigned int dummyVar, bool isInSpan, Matrix<Field>& matrix, ActingElement<Field, LINEAR>* const transporter);

    /// Minimize the last fixed column in the case that the projection is a free module and that the next column is independent.
    void minimizeIndependentK0(Matrix<Field>& matrix, ActingElement<Field, LINEAR>* const transporter);

    /// Minimize the last fixed column in the case that the projection is a free module and that the next column is not independent.
    void minimizeDependentK0(Matrix<Field>& matrix, ActingElement<Field, LINEAR>* const transporter);

    /// Get generators of the inner stabilizer group.
    /// \arg \c matrix The matrix is needed for the definition of the generators.
    /// \arg \c gens The generators will be appended to gens.
    /// \returns The order of the inner stabilizer subgroup.
    My_Long_Int getAutom(const Matrix<Field>& matrix, std::vector<ActingElement<Field, LINEAR> >& gens) const;

    void fillActingElement(unsigned int k0, const Matrix<Field>& root,  ActingElement<Field, LINEAR>& g, const std::vector<unsigned int>& to_fixSeq);

    friend std::ostream& operator<<(std::ostream& os, const InnerStab s)
    {
        os<<s.m_rank<<std::endl;
        os<<s.m_row_partition<<std::endl;
        return os;
    }
};

template<>
struct InnerStab<Field, PLUSMINUSMONOMIAL> : public InnerStab<Field, LINEAR>
{
    InnerStab(void) : InnerStab<Field, LINEAR>() {}

    /// Set the inner stabilizer to the group \f$ GL_k(R) \times {R^*}^n\f$.
    InnerStab(unsigned int k_0, unsigned int n) : InnerStab<Field, LINEAR>(k_0,n) {}

    /// Copy Constructor.
    InnerStab(const InnerStab& rhs) :  InnerStab<Field, LINEAR>(rhs) {}

    void minimize(unsigned int dummyVar, bool isInSpan, Matrix<Field>& matrix, ActingElement<Field, PLUSMINUSMONOMIAL>* const transporter);

    /// Minimize the last fixed column in the case that the projection is a free module and that the next column is not independent.
    void minimizeDependentK0(Matrix<Field>& matrix, ActingElement<Field, PLUSMINUSMONOMIAL>* const transporter);

    /// Get generators of the inner stabilizer group.
    /// \arg \c matrix The matrix is needed for the definition of the generators.
    /// \arg \c gens The generators will be appended to gens.
    /// \returns The order of the inner stabilizer subgroup.
    My_Long_Int getAutom(const Matrix<Field>& matrix, std::vector<ActingElement<Field, PLUSMINUSMONOMIAL> >& gens) const;

    void fillActingElement(unsigned int k0, const Matrix<Field>& root,  ActingElement<Field, PLUSMINUSMONOMIAL>& g, const std::vector<unsigned int>& to_fixSeq);

    friend std::ostream& operator<<(std::ostream& os, const InnerStab s)
    {
        os<<((InnerStab<Field, LINEAR>&) s)<<std::endl;
        return os;
    }
};


template<>
struct InnerStab<Field, SEMILINEAR> : public InnerStab<Field, LINEAR>
{
public:
    unsigned int m_frob_pow;

    InnerStab(void) : InnerStab<Field, LINEAR>(), m_frob_pow(Field::get_r()!=1) {};
    /// Set the inner stabilizer to the group \f$ GL_k(R) \times {R^*}^n\f$.
    InnerStab(unsigned int k, unsigned int n)  :  InnerStab<Field, LINEAR>(k, n), m_frob_pow(Field::get_r()!=1) {}

    /// Minimize the last fixed column.
    void minimize(unsigned int dummyVar, bool isInSpan, Matrix<Field>& matrix, ActingElement<Field, SEMILINEAR>* const transporter);

    /// Minimize the last fixed column in the case that the projection is a free module and that the next column is independent.
    void minimizeIndependentK0(Matrix<Field>& matrix, ActingElement<Field, SEMILINEAR>* const transporter);

    /// Minimize the last fixed column in the case that the projection is a free module and that the next column is not independent.
    void minimizeDependentK0(Matrix<Field>& matrix, ActingElement<Field, SEMILINEAR>* const transporter);

    /// Get generators of the inner stabilizer group.
    /// \arg \c matrix The matrix is needed for the definition of the generators.
    /// \arg \c gens The generators will be appended to gens.
    /// \returns The order of the inner stabilizer subgroup.
    My_Long_Int getAutom(const Matrix<Field>& matrix, std::vector<ActingElement<Field, SEMILINEAR> >& gens) const;

    void fillActingElement(unsigned int k0, const Matrix<Field>& root,  ActingElement<Field, SEMILINEAR>& g,
            const std::vector<unsigned int>& to_fixSeq);

    friend std::ostream& operator<<(std::ostream& os, const InnerStab s)
    {
        os<<((InnerStab<Field, LINEAR>&) s)<<std::endl;
        os<<s.m_frob_pow<<std::endl;
        return os;
    }
};

template<typename RING>
struct InnerStab<RING, LINEAR>
{
    unsigned int* m_type;							        	///< Tells you the type of the projection.
    unsigned int m_rank;							        	///< The sum of the array m_type.
    UnordPartition m_row_partition;			        	///< Sets of rows, which have to be mulitplied by the same unit.
    std::vector<DenseVec<BinaryField> > m_matrixRep;	        	///< The generators expressed as subgroup of \f$ \mathbb{Z}_2^x \f$
    std::vector<unsigned int> m_fixedCols;			///< Tells the order in which the columns were fixed.
    std::vector<bool> m_isFixedCol;						///< Index i is true, if this column is fixed.

    void init(void)
    {
        if(m_type == 0)
        {
            m_type = new unsigned int[2];
            for(unsigned int i=0; i<2; ++i)
            {
                m_type[i]=0;
            }
        }
    }

    /// Default constructor.
    InnerStab(void) : m_type(0), m_rank(0)
    {
        init();
    }

    /// Set the inner stabilizer to the group \f$ GL_k(R) \times {R^*}^n\f$.
    InnerStab(unsigned int k_0, unsigned int n) : m_type(0), m_rank(0), m_row_partition(k_0), m_isFixedCol(n, false)
    {
        init();
    }

    /// Copy Constructor.
    InnerStab(const InnerStab& rhs) : m_rank(rhs.m_rank), m_row_partition(rhs.m_row_partition), m_matrixRep(rhs.m_matrixRep), m_fixedCols(rhs.m_fixedCols), m_isFixedCol( rhs.m_isFixedCol)
    {
        m_type = new unsigned int[2];
        for(unsigned int i=0; i<2; ++i)
        {
            m_type[i]=rhs.m_type[i];
        }
    }

    ~InnerStab(void)
    {
        delete[] m_type;
    }

    /// Assignment
    InnerStab& operator=(const InnerStab& rhs)
    {
        m_rank = rhs.m_rank;
        m_row_partition = rhs.m_row_partition;
        m_matrixRep = rhs.m_matrixRep;
        m_fixedCols = rhs.m_fixedCols;
        m_isFixedCol = rhs.m_isFixedCol;
        for(unsigned int i=0; i<2; ++i)
        {
            m_type[i]=rhs.m_type[i];
        }
        return *this;
    }

    bool needUnit(unsigned int k0) const { return m_type[0]<k0; }

    /// \brief Transform the generators described by \a m_s0, \a m_s1 and \a row_partition to \a m_matrixRep.
    void toMatrixRepresentation(const Matrix<RING>& matrix);

    /// \brief We add the generators, which arise if we stabilize a column which defines a pivot element equal to 2.
    void updateMatrixRepresentation(const Matrix<RING>& matrix);

    /// \brief We apply the group element \a min_g to \a matrix.
    /// If \a transporter is not NULL, we also multiply \a min_g to \a transporter.
    /// not implemented for chainlength==1
    void apply(Matrix<RING>& matrix, const DenseVec<BinaryField>& min_g,  ActingElement<RING, LINEAR>* const transporter);

    /// Minimize the last fixed column.
    void minimize(unsigned int k0, bool isInSpan, Matrix<RING>& matrix, ActingElement<RING, LINEAR>* const transporter);

    /// Minimize the last fixed column in the case that the projection is a free module and that the next column is independent.
    void minimizeIndependentK0(Matrix<RING>& matrix, ActingElement<RING, LINEAR>* const transporter);

    /// Minimize the last fixed column in the case that the projection is not a free module and that the next column is independent.
    void minimizeIndependentK1(Matrix<RING>& matrix, ActingElement<RING, LINEAR>* const transporter);

    /// Minimize the last fixed column in the case that the projection is a free module and that the next column is not independent.
    void minimizeDependentK0(Matrix<RING>& matrix, ActingElement<RING, LINEAR>* const transporter);

    /// Minimize the last fixed column in the case that the projection is not a free module and that the next column is not independent.
    void minimizeDependentK1(Matrix<RING>& matrix, ActingElement<RING, LINEAR>* const transporter);

    /// Get generators of the inner stabilizer group.
    /// \arg \c matrix The matrix is needed for the definition of the generators.
    /// \arg \c gens The generators will be appended to gens.
    /// \returns The order of the inner stabilizer subgroup.
    My_Long_Int getAutom(const Matrix<RING>& matrix, std::vector<ActingElement<RING, LINEAR> >& gens) const;

    void fillActingElement(unsigned int k0, const Matrix<RING>& root,  ActingElement<RING, LINEAR>& g, const std::vector<unsigned int>& to_fixSeq);

    friend std::ostream& operator<<(std::ostream& os, const InnerStab s)
    {
        os<<"not implemented yet"<<std::endl;
        return os;
    }
};


template<typename RING>
void InnerStab<RING, PERMUTATIONAL>::minimize(unsigned int k0, bool isInSpan, Matrix<RING>& matrix,
        ActingElement<RING, PERMUTATIONAL>* const transporter)
{
    if(!isInSpan)
    {
        if (transporter == NULL)
        {
            matrix.gaussElStep(m_rank, m_fixedCols.back(), true, NULL);
        }
        else
        {
            matrix.gaussElStep(m_rank, m_fixedCols.back(), true, &transporter->m_A);
        }
        ++m_rank;
    }
}

//template<typename RING>
//void InnerStab<RING, PERMUTATIONAL>::minimizeIndependentK0(Matrix<RING>& matrix,
//        ActingElement<RING, PERMUTATIONAL>* const transporter)
//{
//    if (transporter == NULL)
//    {
//        matrix.gaussElStep(m_rank, m_fixedCols.back(), true, NULL);
//    }
//    else
//    {
//        matrix.gaussElStep(m_rank, m_fixedCols.back(), true, &transporter->m_A);
//    }
//    ++m_type[0];
//    ++m_rank;
//}

template<typename RING>
void InnerStab<RING, PERMUTATIONAL>::fillActingElement(unsigned int k0, const Matrix<RING>& root, ActingElement<
        RING, PERMUTATIONAL>& g, const std::vector<unsigned int>& to_fixSeq)
{
    unsigned int n = root.getCols();
    Matrix<RING> copy(root);
    copy.permuteCols(g.m_pi);
    std::vector<unsigned int>::const_iterator tofixIter = to_fixSeq.begin();

    for (unsigned int i = 0; i < n; ++i, ++tofixIter)
    {
        m_fixedCols.push_back(*tofixIter);
        m_isFixedCol[*tofixIter] = true;
        unsigned int height = (copy.getColumn(*tofixIter).height(m_rank)).first;
        minimize(k0, (height == RING::chainlength()), copy, &g);
    }
}


template<typename RING>
void InnerStab<RING, LINEAR>::minimizeIndependentK0(Matrix<RING>& matrix, ActingElement<RING, LINEAR>* const transporter)
{
    if(transporter==NULL)
    {
        matrix.gaussElStep(m_rank, m_fixedCols.back(), true, NULL);
    }
    else
    {
        matrix.gaussElStep(m_rank, m_fixedCols.back(), true, &transporter->m_A);
    }
    ++m_rank;
    ++m_type[0];
}

template<typename RING>
void InnerStab<RING,  LINEAR>::minimizeIndependentK1(Matrix<RING>& matrix, ActingElement<RING, LINEAR>* const transporter)
{
    if(m_type[1] == 0)
    {
        // this is the first time the projection does not define a free module, change the representation
        toMatrixRepresentation(matrix);
    }
    if(transporter==NULL)
    {
        matrix.gaussElStep(m_rank, m_fixedCols.back(), true, NULL);
    }
    else
    {
        matrix.gaussElStep(m_rank, m_fixedCols.back(), true, &transporter->m_A);
    }
    updateMatrixRepresentation(matrix);
    ++m_type[1];
    ++m_rank;
}

template<typename RING>
void InnerStab<RING, LINEAR>::minimizeDependentK0(Matrix<RING>& matrix, ActingElement<RING, LINEAR>* const transporter)
{
    if(transporter == NULL)
    {
        matrix.minimizeByColMult(false, m_rank, m_row_partition, m_fixedCols, NULL, NULL);
    }
    else
    {
        matrix.minimizeByColMult(false, m_rank, m_row_partition, m_fixedCols, &transporter->m_A, &transporter->m_phi);
    }
}



template<typename RING>
void InnerStab<RING, LINEAR>::minimizeDependentK1(Matrix<RING>& matrix, ActingElement<RING, LINEAR>* const transporter)
{
    unsigned int k(matrix.getRows()), n(matrix.getCols());
    unsigned int col_pos=m_fixedCols.back();
    const Vec<RING>& column(matrix.getColumn(col_pos));
    unsigned int phi_pos= m_type[0]*(k-m_type[0]) + m_type[0] + col_pos;
    unsigned int size = n + m_type[0]*(k-m_type[0]) + m_type[0];

    // add the column multiplication
    if(column.height().first==0)
    {// otherwise the multiplication is "trivial"
        DenseVec<BinaryField> newRow(size);
        newRow.setBit(phi_pos  );
        m_matrixRep.push_back(newRow);
    }

    typename Vec<RING>::const_reverse_iterator columnIter = column.rbegin()+(k-m_type[0]);
    typename Vec<RING>::const_reverse_iterator columnIterEnd = column.rend();
    unsigned int row=m_type[0]-1;
    for(; columnIter<columnIterEnd; ++columnIter, row--)
    {
        DenseVec<BinaryField> x(size);// which entries may change the value
        if(columnIter->isUnit())
        { // any multiplication changes the value!
            x.setBit(row);
            x.setBit(phi_pos);
        }
        typename Vec<RING>::const_iterator addIter = column.begin()+m_type[0];
        typename Vec<RING>::const_iterator addIterEnd = column.begin()+m_rank;
        unsigned int add_pos=0;
        for( ; addIter<addIterEnd; ++addIter, ++add_pos)
        {
            if(!addIter->isZero()) // then it must be a 2
            {
                x.setBit( m_type[0] +  add_pos*m_type[0] + row);
            }
        }

        //		std::cout<<x<<std::endl;

        std::vector<DenseVec<BinaryField> >::iterator matrixRepIter(m_matrixRep.begin());
        std::vector<DenseVec<BinaryField> >::iterator matrixRepIterEnd(m_matrixRep.end());
        std::vector<unsigned int> changePos;
        unsigned int counter=0;

        for(; matrixRepIter<matrixRepIterEnd; ++matrixRepIter, ++counter)
        {
            if(matrixRepIter->scalar(x))
            {
                changePos.push_back(counter);
            }
        }

        if(!changePos.empty())
        {
            std::vector< DenseVec<BinaryField> >::iterator remStab = m_matrixRep.begin() + changePos.back();
            for(unsigned int i=0; i<changePos.size()-1; ++i)
            {
                m_matrixRep[changePos[i]] += *remStab;
            }
            if( columnIter->isReducable() )
            {
                apply(matrix, *remStab, transporter);
            }
            m_matrixRep.erase( remStab );
        }
    }
}

template<typename RING>
void InnerStab<RING, LINEAR>::toMatrixRepresentation(const Matrix<RING>& matrix)
{
    unsigned int k(matrix.getRows()), n(matrix.getCols());
    unsigned int size = n + m_type[0]*(k-m_type[0]) + m_type[0];
    std::vector<std::vector<unsigned int> > rowSets =  m_row_partition.getSets();
    for(unsigned int i=0; i<rowSets.size(); ++i)
    {
        //		print(std::cout, rowSets[i], "")<<std::endl;

        DenseVec<BinaryField> newRow(size);
        std::vector<unsigned int>::const_iterator setIter(rowSets[i].begin());
        std::vector<unsigned int>::const_iterator setIterEnd(rowSets[i].end());
        for(; setIter<setIterEnd; ++setIter)
        {
            newRow.setBit(*setIter);
        }
        m_matrixRep.push_back(newRow);
    }

    unsigned int phi_offset = m_type[0]*(k-m_type[0]) + m_type[0];
    std::vector<unsigned int>::const_iterator fixedColsIter(m_fixedCols.begin());
    std::vector<unsigned int>::const_iterator fixedColsIterEnd(m_fixedCols.begin()+(m_fixedCols.size()-1));
    for(; fixedColsIter<fixedColsIterEnd; ++fixedColsIter)
    {
        unsigned int index = matrix.getColumn(*fixedColsIter).getFirstUnit().first;
        bool cont_flag = true;
        for(unsigned int i=0; i<rowSets.size() && cont_flag; ++i)
        {
            std::vector<unsigned int>::const_iterator setIter(rowSets[i].begin());
            std::vector<unsigned int>::const_iterator setIterEnd(rowSets[i].end());
            for(; setIter<setIterEnd && cont_flag; ++setIter)
            {
                if(*setIter == index)
                {
                    m_matrixRep[i].setBit(phi_offset+*fixedColsIter);
                    cont_flag = false;
                }
            }
        }
    }

    m_row_partition.clear();
}

template<typename RING>
void InnerStab<RING, LINEAR>::updateMatrixRepresentation(const Matrix<RING>& matrix)
{
    // we assume entry (m_s0+ m_s1 - 1, m_fixedCols.back())  is a pivot element and all entries in this column are reduced!
    unsigned int k(matrix.getRows()), n(matrix.getCols());
    const Vec<RING>& column(matrix.getColumn(m_fixedCols.back()));
    unsigned int phi_pos= m_type[0]*(k-m_type[0]) + m_type[0] + m_fixedCols.back();
    unsigned int U_col_offset = m_type[0] + m_type[1]*m_type[0];
    unsigned int size = n + m_type[0]*(k-m_type[0]) + m_type[0];

    // add the column multiplication
    if((column.height()).first==0)
    { // otherwise the multiplication is "trivial"
        DenseVec<BinaryField> newRow(size);
        newRow.setBit(phi_pos);
        m_matrixRep.push_back(newRow);
    }

    std::vector<DenseVec<BinaryField> >::iterator matrixRepIter(m_matrixRep.begin());
    std::vector<DenseVec<BinaryField> >::iterator matrixRepIterEnd(m_matrixRep.end());
    for(; matrixRepIter<matrixRepIterEnd; ++matrixRepIter)
    {
        typename Vec<RING>::const_iterator colIter(column.begin());
        for(unsigned int row=0; row<m_type[0]; ++row, ++colIter)
        {
            if( column[row].isUnit() && ( matrixRepIter->get(row) ^  matrixRepIter->get(phi_pos) ) )
            {
                matrixRepIter->setBit( U_col_offset + row);
            }
        }
    }
}


template<typename RING>
void InnerStab<RING, LINEAR>::apply(Matrix<RING>& matrix, const DenseVec<BinaryField>& min_g, ActingElement<RING, LINEAR>* const transporter)
{
    std::vector<unsigned int> row_mult;
    for(unsigned int i=0; i<m_type[0]; ++i)
    {
        if(min_g.get(i))
        {
            row_mult.push_back(i);
        }
    }
    matrix.multiplyRows(row_mult, Element<RING>(3), std::vector<unsigned int>());
    if(transporter != NULL)
    {
        transporter->m_A.multiplyRows(row_mult, Element<RING>(3), std::vector<unsigned int>());
    }
    for(unsigned int i=0; i<m_type[0]; ++i)
    {
        for(unsigned int j=0; j<m_type[1]; ++j)
        {
            if(min_g.get(m_type[0] + i + j*m_type[0]))
            {
                matrix.addRows(m_type[0]+j, i);
                if(transporter!=0)
                {
                    transporter->m_A.addRows(m_type[0]+j, i);
                }
            }
        }
    }

    unsigned int phi_beg = m_type[0]*(matrix.getRows()-m_type[0]) + m_type[0] ;
    unsigned int end = phi_beg + matrix.getCols();
    for(unsigned int i=phi_beg; i<end; ++i)
    {
        if(min_g.get(i))
        {
            matrix.multiplyColumn(i - phi_beg,  Element<RING>(3) );
            if(transporter!=NULL)
            {
                transporter->m_phi[i-phi_beg] *= 3;
            }
        }
    }
}


template<typename RING>
My_Long_Int InnerStab<RING, LINEAR>::getAutom(const Matrix<RING>& matrix, std::vector<ActingElement<RING, LINEAR> >& gens) const
{
    unsigned int n = matrix.getCols();
    unsigned int k = matrix.getRows();

    const std::vector<Vec<RING> >& matEntries = matrix.getEntries();
    std::vector<unsigned int> unit_pos;
    for(unsigned int i=0; i<n; ++i)
    {
        unit_pos.push_back( matEntries[i].getFirstUnit().first);
    }

    if( m_type[1] == 0)
    {
        std::vector<std::vector<unsigned int> > sets = m_row_partition.getSets();
        for(unsigned int setIndex=0; setIndex<sets.size(); ++setIndex)
        {
            const std::vector<unsigned int>& this_set = sets[setIndex];

            Matrix<RING> A;
            A.identity(k);
            Vec<RING> phi(n,1);

            for(unsigned int index=0; index<this_set.size(); ++index)
            {
                unsigned int my_pos = this_set[index];
                A.setElement( my_pos, my_pos,  3);
            }

            std::vector<unsigned int>::const_iterator this_setIterBeg = this_set.begin();
            std::vector<unsigned int>::const_iterator  this_setIterEnd = this_set.end();
            for(unsigned int i=0; i<n; ++i)
            {
                std::vector<unsigned int>::const_iterator meetIter = find(this_setIterBeg, this_setIterEnd, unit_pos[i]);
                if(meetIter != this_setIterEnd)
                {
                    phi[i] = 3;
                }
            }
            gens.push_back( ActingElement<RING, LINEAR>(A , phi)  );
        }
        return ((My_Long_Int) 1)<<sets.size();
    }

    // we find the data in m_matrixRep;
    My_Long_Int res(1);
    std::vector<DenseVec<BinaryField> >::const_iterator matrixRepIter = m_matrixRep.begin();
    std::vector<DenseVec<BinaryField> >::const_iterator matrixRepIterEnd = m_matrixRep.end();
    for( ; matrixRepIter<matrixRepIterEnd; ++matrixRepIter)
    {
        std::vector<Vec<RING> > AEntries(k, Vec<RING>(k, 0) );
        for(unsigned int i=0; i<m_type[0]; ++i)
        {
            if(matrixRepIter->get(i))
            {
                AEntries[i][i] = 3;
            }
            else
            {
                AEntries[i][i] = 1;
            }
        }
        for(unsigned int i=m_type[0]; i<k; ++i)
        {
            AEntries[i][i] = 1;
        }

        for(unsigned int i=0; i<m_type[0]; ++i)
        {
            for(unsigned int j=0; j<m_type[1]; ++j)
            {
                if(matrixRepIter->get(m_type[0] + i + j*m_type[0]))
                {
                    AEntries[m_type[0]+j][i] = 1;
                }
            }
        }
        Vec<RING> phi(n,1);
        unsigned int phi_offset= m_type[0]*(m_type[1]) + m_type[0];
        bool do_multOfOrder = false;
        for(unsigned int i=0; i<n; ++i)
        {
            if(matrixRepIter->get(phi_offset+i))
            {
                phi[i] = 3;
                do_multOfOrder = true;
            }
        }

        gens.push_back( ActingElement<RING, LINEAR>(Matrix<RING>(AEntries) , phi)  );

        if(do_multOfOrder)
        {
            res <<=1;
        }
    }
    return res;
}


template<typename RING>
void InnerStab<RING, LINEAR>::minimize(unsigned int k0, bool isInSpan, Matrix<RING>& matrix, ActingElement<RING, LINEAR>* const transporter)
{
    if(!isInSpan)
    {
        if(m_type[0]<k0)
        {
            minimizeIndependentK0(matrix, transporter);
        }
        else
        {
            minimizeIndependentK1(matrix, transporter);
        }
    }
    else
    {
        if(m_type[1]==0)
        {
            minimizeDependentK0(matrix, transporter);
        }
        else
        {
            minimizeDependentK1(matrix, transporter);
        }
    }
}

template<typename RING>
void InnerStab<RING, LINEAR>::fillActingElement(unsigned int k0, const Matrix<RING>& root,  ActingElement<RING, LINEAR>& g, const std::vector<unsigned int>& to_fixSeq)
{
    unsigned int n = root.getCols();
    Matrix<RING> copy(root);
    copy.permuteCols(g.m_pi);
    std::vector<unsigned int>::const_iterator tofixIter = to_fixSeq.begin();
    for(unsigned int i=0; i<n; ++i, ++tofixIter)
    {
        m_fixedCols.push_back(*tofixIter);
        m_isFixedCol[*tofixIter] = true;
        unsigned int height = (copy.getColumn(*tofixIter).height(m_rank)).first;
        minimize(k0, (height==RING::chainlength()), copy, &g);
    }
}

#endif /*INNER_STAB_HPP_*/
