/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  weights.hpp
///  \brief Defines the symmetric weight.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#ifndef WEIGHTS_HPP_
#define WEIGHTS_HPP_

#include <iostream>
#include <map>
#include "vector.hpp"
#include "codewordEnumerator.hpp"

template<typename RING>
class Matrix;

template<unsigned int chainlength>
struct SymWeight
{
	unsigned int* qty;
	SymWeight() : qty(new unsigned int[chainlength]) {
	    for(unsigned int i=0; i<chainlength; i++)
	    {
	        qty[i] = 0;
	    }
	}
	~SymWeight()
	{
		delete[] qty;
	}
	SymWeight(const SymWeight& rhs) : qty(new unsigned int[chainlength])
	{
		for(unsigned int i=0; i<chainlength; i++)
		{
			qty[i] = rhs.qty[i];
		}
	}
	template<typename RING>
	SymWeight(const Vec<RING>& v) : qty(new unsigned int[chainlength])
	{
		v.getWeight();
	}

	void operator=(const SymWeight& rhs)
	{
		for(unsigned int i=0; i<chainlength; i++)
		{
			qty[i] = rhs.qty[i];
		}
	}

	void reset(void)
	{
		for(unsigned int i=0; i<chainlength; i++)
		{
			qty[i] = 0;
		}
	}

	inline void add(unsigned int height, unsigned int n=1);

	int cmp(const SymWeight& w) const;
	bool operator<(const SymWeight& w) const;
	bool operator<=(const SymWeight& w) const;

	bool operator!=(const SymWeight& w) const;
	SymWeight successor(void) const;


	static SymWeight max(unsigned int len)
	{
	    return SymWeight(); // all zero is the heighest possible weight
	}

    static SymWeight minNonzero(void)
    {
        SymWeight res;
        res.qty[0] = 1;
        return res;
    }

	unsigned int HammingWeight() const
	{
	    return qty[0]  + ( qty[1]<<1);
	}

//	template<typename RING>
//	void getWeight(const Codeword<RING>& v)
//	{
//		reset();
//		typename Codeword<RING>::const_iterator iter(v.begin());
//		typename Codeword<RING>::const_iterator iterEnd(v.end());
//
//		for(; iter<iterEnd; iter++)
//		{
//			add(iter->height());
//		}
//	}
//
//	void getWeight(const Codeword<Z4>& v)
//	{
//	    *this = v.getWeight();
//	}
};

template<>
struct SymWeight<1u>
{
	unsigned int qty;
	SymWeight(unsigned int i=0) : qty(i) {}
	SymWeight(const SymWeight& rhs) : qty(rhs.qty){}

	template<typename RING>
	SymWeight(const Vec<RING>& v) : qty(v.getWeight().qty)
	{
	}

	void operator=(const SymWeight& rhs)
	{
		qty = rhs.qty;
	}

	void add(unsigned int height)
	{
	    qty += (height == 0);
	}
	bool operator<(const SymWeight& w) const { return qty<w.qty; }
	bool operator<=(const SymWeight& w) const { return qty<=w.qty; }
	bool operator!=(const SymWeight& w) const { return qty != w.qty; }
	bool operator==(const SymWeight& w) const { return qty == w.qty; }

	SymWeight<1> successor(void) const
	{
	    SymWeight<1> res(*this);
	    ++res.qty;
	    return res;
	}

    static SymWeight<1u> max(unsigned int len)
    {
        return SymWeight<1u>(len);
    }

    static SymWeight<1u> minNonzero(void)
    {
        return SymWeight<1u>(1);
    }

    unsigned int HammingWeight() const
    {
        return qty;
    }
};

template<unsigned int chainlength>
std::ostream& operator<<(std::ostream& os, const SymWeight<chainlength>& w)
{
	os<<"["<<w.qty[0];
	for(unsigned int i=1; i<chainlength; i++)
	{
		os<<", "<<w.qty[i];
	}
	os<<"]";
	return os;
}

template<>
std::ostream& operator<<(std::ostream& os, const SymWeight<1u>& w);


template<typename WEIGHT>
class Distribution
{
private:
	std::map<WEIGHT, unsigned int> map;
public:
	Distribution(void) {};
	Distribution(const Distribution& symWD) : map(symWD.map){}

	template<typename RING>
	Distribution(const Matrix<RING>& mat);

//	template<typename RING>
//	Distribution(const Matrix<RING>& mat, unsigned int min_d);

	inline void add(const WEIGHT& w, unsigned int qty=1);

	template<typename W>
	friend std::ostream& operator<<(std::ostream& os, const Distribution<W>& wd);

	unsigned int getMinWeight(void) const;
	Distribution getDual(unsigned int n, unsigned int k) const;
	unsigned int getDualMinDist(unsigned int n) const;

	typename std::map<WEIGHT, unsigned int>::const_iterator begin(void) const { return map.begin(); }
	typename std::map<WEIGHT, unsigned int>::const_iterator end(void) const { return map.end(); }

	Distribution<SymWeight<1u> > GrayImageDistribution(void) const;

	typename std::map<WEIGHT, unsigned int>::const_iterator lower_bound(const WEIGHT& w) const
	{
	    return map.lower_bound(w);
	}

	bool operator<(const Distribution& rhs) const
	{
	    typename std::map<WEIGHT, unsigned int>::const_iterator iter(this->map.begin());
	    typename std::map< WEIGHT, unsigned int>::const_iterator iterEnd(this->map.end());
	    typename std::map< WEIGHT, unsigned int>::const_iterator rhsIter(rhs.map.begin());
	    typename std::map<WEIGHT, unsigned int>::const_iterator rhsIterEnd(rhs.map.end());

	    for( ; iter!=iterEnd && rhsIter!=rhsIterEnd; ++iter, ++rhsIter)
	    {
	        if(iter->first < rhsIter->first)
	        {
	            return true;
	        }
	        else if(rhsIter->first < iter->first)
	        {
	            return false;
	        }
            if(iter->second < rhsIter->second)
            {
                return true;
            }
            else if(iter->second > rhsIter->second)
            {
                return false;
            }
	    }

	    if(iter!=iterEnd)
	    {
	        return false;
	    }
	    else
	    {
	        if(rhsIter!=rhsIterEnd)
	        {
	            return true;
	        }
	    }
	    return false;
	}

//	bool empty(void) const
//	{
//	    return map.empty();
//	}
};

template<>
unsigned int Distribution<SymWeight<1u> >::getMinWeight(void) const;

template<>
Distribution<SymWeight<1u> > Distribution<SymWeight<1u> >::getDual(unsigned int n, unsigned int k) const;

template<>
unsigned int Distribution<SymWeight<1u> >::getDualMinDist(unsigned int n) const;

template<>
unsigned int Distribution<SymWeight<2u> >::getMinWeight(void) const;

template<typename WEIGHT>
Distribution<SymWeight<1u> > Distribution<WEIGHT>::GrayImageDistribution(void) const
{
    std::map<SymWeight<2u>, unsigned int>::const_iterator wdIter(begin());
    std::map<SymWeight<2u>, unsigned int>::const_iterator wdIterEnd(end());

    Distribution<SymWeight<1u> > result;

    for(; wdIter!=wdIterEnd; ++wdIter)
    {
        result.add( wdIter->first.qty[0] + 2*wdIter->first.qty[1] , wdIter->second);
    }
    return result;
}

//template<>
//template<>
//SymWeightDist<1u>::SymWeightDist(const Matrix<Field>& mat);
//
//template<>
//template<>
//SymWeightDist<1u>::SymWeightDist(const Matrix<BinaryField>& mat);
//
//template<>
//template<>
//SymWeightDist<2u>::SymWeightDist(const Matrix<Z4>& mat);
//
//template<>
//template<>
//SymWeightDist<2u>::SymWeightDist(const Matrix<F2_F2>& mat);

template<unsigned int chainlength>
SymWeight<chainlength> SymWeight<chainlength>::successor(void) const
{
    SymWeight<chainlength> res(*this);
    ++res.qty[chainlength-1];
    return res;
}





template<unsigned int chainlength>
bool SymWeight<chainlength>::operator<(const SymWeight<chainlength>& w) const
{
    return cmp(w)<0;
}

template<unsigned int chainlength>
bool SymWeight<chainlength>::operator<=(const SymWeight<chainlength>& w) const
{
    return cmp(w)<=0;
}

template<unsigned int chainlength>
int SymWeight<chainlength>::cmp(const SymWeight& w) const
{
	for(unsigned int i=0; i<chainlength; ++i)
	{
		if(qty[i]<w.qty[i])
		{
		    if(qty[i] == 0)
		    {
		        return 1;
		    }
			return -1;
		}
		if(qty[i]>w.qty[i])
		{
		    if(w.qty[i] == 0)
		    {
		        return -1;
		    }
			return 1;
		}
	}
	return 0;
}

template<unsigned int chainlength>
bool SymWeight<chainlength>::operator!=(const SymWeight& w) const
{
	for(unsigned int i=0; i<chainlength; ++i)
	{
		if(qty[i] != w.qty[i])
		{
			return true;
		}
	}
	return false;
}

template<unsigned int chainlength>
void SymWeight<chainlength>::add(unsigned int h,  unsigned int n)
{
	if( h != chainlength)
	{
		qty[h]+=n;
	}
}

template<typename WEIGHT>
std::ostream& operator<<(std::ostream& os, const Distribution<WEIGHT>& wd)
{
	typename std::map<WEIGHT, unsigned int>::const_iterator iter=wd.map.begin();
	typename std::map<WEIGHT, unsigned int>::const_iterator iterEnd=wd.map.end();

	if(iter != iterEnd)
	{
		os<<iter->second<<"*"<<iter->first;
		++iter;
	}

	while(iter != iterEnd)
	{
		os<<" + "<<iter->second<<"*"<<iter->first;
		++iter;
	}
	return os;
}

template<typename WEIGHT>
void Distribution<WEIGHT>::add(const WEIGHT& w, unsigned int qty)
{
	typename std::map<WEIGHT, unsigned int>::iterator iter=map.find(w);
	if( iter == map.end())
	{
		map.insert(std::pair<WEIGHT, unsigned int>(w, qty));
	}
	else
	{
		iter->second += qty;
	}
}

//template<>
//SymWeightDist<1> SymWeightDist<2>::toLeeWeightDist(void) const;

//template<>
//template<>
//SymWeightDist<1u>::SymWeightDist(const Matrix<Field>& mat)
//{
//    std::vector<unsigned int> type(1, mat.getRows());
//    CodewordEnumerator<Field> enumerator(type, mat);
//    do
//    {
//        add( enumerator->getWeight() );
//    }while(++enumerator);
//}
//
//template<>
//template<>
//SymWeightDist<1u>::SymWeightDist(const Matrix<BinaryField>& mat)
//{
//    std::vector<unsigned int> type(1, mat.getRows());
//    CodewordEnumerator<BinaryField> enumerator(type, mat);
//    do
//    {
//        add( enumerator->getWeight() );
//    }while(++enumerator);
//}
//
template<typename WEIGHT>
template<typename RING>
Distribution<WEIGHT>::Distribution(const Matrix<RING>& mat)
{
    Matrix<RING> copy( mat );
    std::vector<unsigned int> type = copy.toOrderedBasis(false, NULL, NULL);
    if(RING::chainlength() > 1)
    {
        type[1] = copy.getRows() - type[0];
    }
    CodewordEnumerator<RING> enumerator(type, copy);
    do
    {
        add( enumerator->getWeight() );
    }while(++enumerator);
}

//template<typename WEIGHT>
//template<typename RING>
//Distribution<WEIGHT>::Distribution(const Matrix<RING>& mat, unsigned int min_d)
//{
//    Matrix<RING> copy( mat );
//    std::vector<unsigned int> type = copy.toOrderedBasis(false, NULL, NULL);
//    if(RING::chainlength() > 1)
//    {
//        type[1] = copy.getRows() - type[0];
//    }
//    CodewordEnumerator<RING> enumerator(type, copy);
//    add(WEIGHT());
//    WEIGHT w;
//
//    while(++enumerator)
//    {
//        w = enumerator->getWeight();
//        if(w.HammingWeight() < min_d)
//        {
//            this->map.clear();
//            return;
//        }
//        add( w );
//    }
//}



//
//template<>
//template<>
//SymWeightDist<2u>::SymWeightDist(const Matrix<F2_F2>& mat)
//{
//    std::vector<unsigned int> type(2, 0);
//    Matrix<F2_F2> copy( mat );
//    type[0] = copy.toOrderedBasis(false, NULL, NULL);
//    type[1] = copy.getRows() - type[0];
//    CodewordEnumerator<F2_F2> enumerator(type, copy);
//
//    do
//    {
//        add( enumerator->getWeight() );
//    }while(++enumerator);
//}



#endif /*WEIGHTS_HPP_*/
